!cpu 6510

; ______________________________________________________________________
!source "colorcodes.asm"
; ______________________________________________________________________
; ____________ ZEROPAGE ADRESSES

fw_savea              = $02
fw_savex              = $03
fw_savey              = $04
; ______________________________
;         $05 - $DF free to use!
; ______________________________
; $E0 - (max.) $F$  KRILL LOADER
; ______________________________
fw_color_d020         = $f4
fw_color_d021         = $f5

fw_irq_ready          = $f6
fw_d012_value         = $f7
fw_d011_value         = $f8
fw_irq_lo             = $f9
fw_irq_hi             = $fa

fw_global_ct          = $fb             ; $fb, $fc, $fd

; music player uses:    $fe, $ff

; ____________ RAM ADRESSES
fw_code_start         = $0200
fw_irq_addr           = $0300
fw_music_init         = $0c00
fw_music_play         = fw_music_init+3
fw_music_volume       = fw_music_init+$135
fw_part_start_lo      = $1ffe
fw_part_start_hi      = $1fff

; ____________ KRILL
  !source "loadersymbols-c64.inc"
; ______________________________________________________________________
; ______________________________________________ INCLUDES
*= fw_music_init
  !bin "music/dt2remix-noedit-0c00.sid",,$7e
fw_music_end:

!if RELEASE = 1 {
  *= loadcompd
    !bin "loader-c64.c64",,2
  *= $2000
    !bin "install-c64.c64",,2
}

!if RELEASE = 0 {
  *= $0400
}

fw_krill_end:
; ______________________________________________________________________
; ______________________________________________ CODE START
  jmp fw_basic_fade
fw_reloc_code_01_start:
!pseudopc fw_code_start {
  lda #<fw_nmi
  sta $fffa
  lda #>fw_nmi
  sta $fffb

  jsr fw_init_vic

  lda #$00
  sta fw_color_d020
  sta fw_color_d021
  sta fw_global_ct
  sta fw_global_ct+1
  sta fw_global_ct+2
  sta fw_d012_value
!if RELEASE = 1 {
  lda fw_d011_value
} else {
  lda #$00
}
  sta fw_d011_value

  lda #$00
  jsr fw_music_init

  bit fw_init_timer

  lda #<fw_irq
  sta $fffe
  sta fw_irq_lo
  lda #>fw_irq
  sta $ffff
  sta fw_irq_hi

  asl $d019
  cli
; ________________________________________________ MAIN LOOP
!zone MAINLOOP
fw_mainloop:

fw_load_flag =*+1
  lda #$00
  bne fw_skip_load

  lda #0
  sta loadaddrlo
  sta loadaddrhi
  ldx #<fw_filename
  ldy #>fw_filename

  jsr loadcompd
  bcc +

  jmp *

+
  inc fw_partnum

fw_skip_load:
  lda fw_part_start_lo
  sta fw_exit+1
  lda fw_part_start_hi
  sta fw_exit+2

fw_exit:
  jmp $0000

fw_back2framework:
  jsr fw_wait_irq
  jsr fw_init_vic

  lda fw_partnum
  jsr fw_hex2screen
  sta fw_filename
  stx fw_filename+1

  jmp fw_mainloop
; ____________
fw_filename:
  !scr "00"
fw_partnum:
  !byte $00
; ____________
fw_wait_irq:
  lda #$00
  sta fw_irq_ready
-
  lda fw_irq_ready
  beq -
  rts
; ______________________________________________ HEX2SCREEN
!zone HEX2SCREEN
fw_hex2screen:
    sta .savea
    and #%00001111
    tax
    lda .hextab,x
    sta .low_nibble
.savea = *+1
    lda #0
    lsr
    lsr
    lsr
    lsr
    tax
    lda .hextab,x           ; high nibble
.low_nibble = *+1
    ldx #0
    rts

.hextab:
    !scr "0123456789abcdef"
; ______________________________________________ INIT VIC
!zone INIT_VIC
fw_init_vic:
  lda $d016
  sta .d016_sav
  lda $d018
  sta .d018_sav

  lxa #$00 ; fill A and X with $00 is safe only with 0!
-
  sta $d000,x
  inx
  cpx #$2f
  bne -

fw_preserve_screen =*+1
!if RELEASE = 1 {
  lda #1
} else {
  lda #0
}
  beq +
  lda #0
  sta fw_preserve_screen
  lda fw_d011_value
  sta $d011
.d016_sav =*+1
  lda #0
  sta $d016
.d018_sav =*+1
  lda #0
  sta $d018
+
  lda #$01
  sta $d01a
  rts
; ______________________________________________ INIT TIMER
!zone INIT_TIMER
fw_init_timer:
  cmp $d012
  bne fw_init_timer
  ldy #$08
  sty $dc04
-
  dey
  bne -
  sty $dc05
  sta $dc0e,y
  lda #$11
  cmp $d012
  sty $d015
  bne fw_init_timer
  rts
}
fw_reloc_code_01_end:
; ______________________________________________ IRQ
!zone IRQ
fw_reloc_code_02_start:
!pseudopc fw_irq_addr {
fw_irq:
  sta fw_savea
  stx fw_savex
  sty fw_savey

  lda fw_color_d020
  sta $d020
  lda fw_color_d021
  sta $d021

  jsr fw_music_play

  clc
  lda fw_global_ct
  adc #$01
  sta fw_global_ct
  lda fw_global_ct+1
  adc #$00
  sta fw_global_ct+1
  lda fw_global_ct+2
  adc #$00
  sta fw_global_ct+2

  lda fw_d012_value
  sta $d012

  lda fw_d011_value
  sta $d011

  lda fw_irq_lo
  sta $fffe
  lda fw_irq_hi
  sta $ffff

  lda #$01
  sta fw_irq_ready

  asl $d019
  lda fw_savea
  ldx fw_savex
  ldy fw_savey
fw_nmi:
  rti
}
fw_reloc_code_02_end:
; ______________________________________________________________________
!zone BASIC_FADE
fw_basic_fade:

  sei
  lda #$7f
  sta $dc0d
  sta $dd0d
  lda $dc0d
  lda $dd0d

; !BASIC FADE TO BE ADDED!
!if RELEASE = 1 {
  lda #$36
  sta $01

  jsr install
}

  lda #$35
  sta $01

; ______________________________________________________________________
!zone COPY_CODE
  ldx #(fw_reloc_code_01_end-fw_reloc_code_01_start)
-
  lda fw_reloc_code_01_start,x
  sta fw_code_start,x
  dex
  cpx #$ff
  bne -

  ldx #(fw_reloc_code_02_end-fw_reloc_code_02_start)
-
  lda fw_reloc_code_02_start,x
  sta fw_irq_addr,x
  dex
  cpx #$ff
  bne -

!if RELEASE = 1 {
  lda #$00
}
!if RELEASE = 0 {
  lda #$01
}
  sta fw_load_flag
!if RELEASE = 1 {
    jsr $3600
} else {
    bit $0000
}
  jmp fw_code_start
fw_endmarker:

!if RELEASE = 1 {
*= $3600
    !bin "starrise.c64",,2
}
