!cpu 6510

!source "colorcodes.asm"
; ______________________________________________________________________
scrollspeed           = $01
enable                = $20
disable               = $2c

; ____________ ZEROPAGE ADRESSES
petscii_bg1           = $05
petscii_bg2           = petscii_bg1+1
screen_pt             = petscii_bg2+1
colram_pt             = screen_pt+2
scroll                = colram_pt+2
jump_loc              = scroll+1
jumptable_pointer     = jump_loc+2
irq_ready             = jumptable_pointer+1
zp_temp               = irq_ready+1

; ____________ RAM ADRESSES
vidmem0               = $2000
vidmem1               = $2400
colram_buffer         = $2800
; ______________________________________________ INCLUDES
!source "../../bin/framework/frameworksymbols.inc"

!if RELEASE = 0 {
  *= $0400
    !bin "../../bin/framework/framework-noloader.prg",,2

  *= fw_music_init
    !bin "../framework/music/dt2remix-part-08.sid",,$7e
}
; ______________________________________________________________________

*= $1ffe
    !byte <code_start
    !byte >code_start

*= vidmem0
    !fi 1000, $20

*= vidmem1
    !fi 1000, $20

*= colram_buffer
    !fi 1000, $00

*= $2c00
screendata:
    !bin "resources/sky-01.c64",1000,$62
    !bin "resources/sky-02.c64",1000,$62
    !bin "resources/sky-03.c64",1000,$62
    !bin "resources/sky-04.c64",1000,$62
    !bin "resources/sky-05.c64",1000,$62
    !bin "resources/sky-06.c64",1000,$62
screen_grey:
    !bin "resources/sky-07.c64",1000,$62
    !bin "resources/sky-08.c64",1000,$62
screen_black:
    !fi 40, 160
    !bin "resources/sky-09.c64",1000,$62
    !bin "resources/sky-10.c64",1000,$62
    !bin "resources/sky-11.c64",1000,$62
    !bin "resources/sky-12.c64",1000,$62
    !bin "resources/sky-13.c64",1000,$62
screendata_end:
    !fi 40, 160
!align 255,0
coldata:
    !bin "resources/sky-01.c64",1000,$44a
    !bin "resources/sky-02.c64",1000,$44a
    !bin "resources/sky-03.c64",1000,$44a
    !bin "resources/sky-04.c64",1000,$44a
    !bin "resources/sky-05.c64",1000,$44a
    !bin "resources/sky-06.c64",1000,$44a
    !bin "resources/sky-07.c64",1000,$44a
    !bin "resources/sky-08.c64",1000,$44a
    !fi 40, 0
    !bin "resources/sky-09.c64",1000,$44a
    !bin "resources/sky-10.c64",1000,$44a
    !bin "resources/sky-11.c64",1000,$44a
    !bin "resources/sky-12.c64",1000,$44a
    !bin "resources/sky-13.c64",1000,$44a
coldata_end:
    !fi 40, 0
; ______________________________________________________________________
; ______________________________________________ IRQs
!fi $100-(<*), 0
irq:
    sta fw_savea
    stx fw_savex
    sty fw_savey

    jsr fake
    jsr fake
    bit $ea
    bit $ea
    nop

    lda petscii_bg1
    sta $d021

    sec
    lda scroll
    sbc #scrollspeed
    cmp #$ff
    bne +
    lda $d018
    eor #($84 XOR $94)
    sta $d018
    lda #$07
+
    sta scroll

    lda fw_d011_value
    and #%11111000
    ora scroll
    sta fw_d011_value

enable_bgswitch:
    bit bgswitch

irq1_line =*+1
    lda #$f7
    sta $d012

    lda fw_d011_value
    sta $d011

    lda #<irq1
    sta $fffe
    lda #>irq1
    sta $ffff

    lda #$01
    sta irq_ready

    asl $d019
    lda fw_savea
    ldx fw_savex
    ldy fw_savey
    rti

fake:
    rts

irq1:
    sta fw_savea
    stx fw_savex
    sty fw_savey

    lda $d012
-
    cmp $d012
    beq -

    lda petscii_bg2
    sta $d021

    lda #$00
    sta $d012

    lda #<fw_irq_addr
    sta $fffe
    lda #>fw_irq_addr
    sta $ffff

    asl $d019
    lda fw_savea
    ldx fw_savex
    ldy fw_savey
    rti

bgswitch:
    dec irq1_line
    lda irq1_line
    cmp #$2f
    bcs +
    lda #<(screen_black+40)
    sta mod_newdat1
    lda #>(screen_black+40)
    sta mod_newdat2
    lda petscii_bg1
    pha
    lda petscii_bg2
    sta petscii_bg1
    sta $d021
    pla
    sta petscii_bg2
    lda #$f7
    sta irq1_line
    lda #disable
    sta enable_bgswitch
+
    rts
; ______________________________________________________________________
; ______________________________________________ CODE START
code_start:

    jsr gen_colram_copy

    lda #$00
    sta fw_color_d020
    sta fw_color_d021
    sta jumptable_pointer
    sta irq_ready
    sta petscii_bg1

    lda #12
    sta petscii_bg2

    lda #$07
    sta scroll

    ldx #0
-
    lda screendata+$000,x
    sta vidmem0+$000,x
    sta vidmem1+$000,x
    lda coldata+$000,x
    sta $d800+$000,x
    lda screendata+$100,x
    sta vidmem0+$100,x
    sta vidmem1+$100,x
    lda coldata+$100,x
    sta $d800+$100,x
    lda screendata+$200,x
    sta vidmem0+$200,x
    sta vidmem1+$200,x
    lda coldata+$200,x
    sta $d800+$200,x
    lda screendata+$2e8,x
    sta vidmem0+$2e8,x
    sta vidmem1+$2e8,x
    lda coldata+$2e8,x
    sta $d800+$2e8,x
    inx
    bne -

    lda #%00000011
    sta $dd00

    lda #$84
    sta $d018

    lda #$08
    sta $d016

    lda #$17
    sta fw_d011_value

    lda #<(screendata+1000)
    sta screen_pt
    lda #>(screendata+1000)
    sta screen_pt+1

    lda screen_pt
    sta colram_pt
    clc
    lda screen_pt+1
    adc #>((coldata+1000)-(screendata+1000))
    sta colram_pt+1

    clc
    lda fw_global_ct
    adc #$08    ;48 ;88
    sta end_time_a
    lda fw_global_ct+1
    adc #$0a    ;09 ;08
    sta end_time_b
    lda fw_global_ct+2
    adc #$00
    sta end_time_c

    jsr fw_wait_irq

    lda #$24
    sta fw_d012_value

    lda #<irq
    sta fw_irq_lo
    lda #>irq
    sta fw_irq_hi
; ______________________________________________________________________
; ______________________________________________ MAINLOOP
mainloop:
    jsr wait_irq

    lda fw_global_ct
end_time_a =*+1
    cmp #0
    bne +
    lda fw_global_ct+1
end_time_b =*+1
    cmp #0
    bne +
    lda fw_global_ct+2
end_time_c =*+1
    cmp #0
    bne +
    jmp exit_part
+
    lda scroll
    cmp #7
    bne +
    jsr colram_copy
    jmp mainloop
+
    cmp #5
    bne +
    jsr prepare_hard
    jmp mainloop
+
    cmp #4
    bne +
    jsr prepare_hard
    jmp mainloop
+
    cmp #3
    bne +
    jsr prepare_hard
    jmp mainloop
+
    cmp #2
    bne +
    jsr prepare_hard
    jmp mainloop
+
    cmp #0
    bne +
    jsr newdata
    jmp mainloop
+
    jmp mainloop
; ______________________________________________________________________
prepare_hard:
    ldx jumptable_pointer
    lda jumptable_lo,x
    sta jump_loc
    lda jumptable_hi,x
    sta jump_loc+1
    inc jumptable_pointer
    lda jumptable_pointer
    cmp #$08
    bne +
    lda #$00
    sta jumptable_pointer
+
    jmp (jump_loc)

newdata:
    nop

    clc
    lda screen_pt
    adc #40
    sta screen_pt
    lda screen_pt+1
    adc #0
    sta screen_pt+1

    lda #<screendata_end
    cmp screen_pt
    bne +
    lda #>screendata_end+1
    cmp screen_pt+1
    bne +

    lda #$60
    sta newdata
+
    lda screen_pt
    sta colram_pt
    clc
    lda screen_pt+1
    adc #>(coldata-screendata)
    sta colram_pt+1

mod_newdat1 =*+1
    lda #<(screen_grey+40)
    cmp screen_pt
    bne +
mod_newdat2 =*+1
    lda #>(screen_grey+40)
    cmp screen_pt+1
    bne +
    lda #enable
    sta enable_bgswitch
+
    rts

wait_irq:
    lda #0
    sta irq_ready
-
    lda irq_ready
    beq -
    rts

exit_part:
    lda #$00
    sta fw_irq_ready
-
    lda fw_irq_ready
    beq -

    lda #<fw_irq
    sta $fffe
    sta fw_irq_lo
    lda #>fw_irq
    sta $ffff
    sta fw_irq_hi

    lda #$00
    sta $d012
    sta fw_d012_value

    lda #$0b
    sta $d011
    sta fw_d011_value
!if RELEASE = 0 {
    jmp *
} else {
    jmp fw_back2framework
}
; ______________________________________________________________________
; ___________________________________________________ HARDSCROLL
!zone HARDSCROLL

jumptable_lo:
    !byte <hardscroll_a, <hardscroll_b
    !byte <colscroll_a, <colscroll_b
    !byte <hardscroll_c, <hardscroll_d
    !byte <colscroll_a, <colscroll_b

jumptable_hi:
    !byte >hardscroll_a, >hardscroll_b
    !byte >colscroll_a, >colscroll_b
    !byte >hardscroll_c, >hardscroll_d
    !byte >colscroll_a, >colscroll_b

hardscroll_a:
    ldy #$27
-
!for i, 0, 11 {
    lda vidmem0+((i+1)*40),y
    sta vidmem1+(i*40),y
}
    dey
    bpl -
    jmp back_jump

hardscroll_b:
    ldy #$27
-
!for i, 12, 23 {
    lda vidmem0+((i+1)*40),y
    sta vidmem1+(i*40),y
}
    lda (screen_pt),y
    sta vidmem1+(24*40),y
    dey
    bpl -
    jmp back_jump

hardscroll_c:
    ldy #$27
-
!for i, 0, 11 {
    lda vidmem1+((i+1)*40),y
    sta vidmem0+(i*40),y
}
    dey
    bpl -
    jmp back_jump

hardscroll_d:
    ldy #$27
-
!for i, 12, 23 {
    lda vidmem1+((i+1)*40),y
    sta vidmem0+(i*40),y
}
    lda (screen_pt),y
    sta vidmem0+(24*40),y
    dey
    bpl -
    jmp back_jump

colscroll_a:
    ldy #$27
-
!for i, 0, 11 {
    lda $d800+((i+1)*40),y
    sta colram_buffer+(i*40),y
}
    dey
    bpl -
    jmp back_jump

colscroll_b:
    ldy #$27
-
!for i, 12, 23 {
    lda $d800+((i+1)*40),y
    sta colram_buffer+(i*40),y
}
    lda (colram_pt),y
    sta colram_buffer+(24*40),y
    dey
    bpl -
    jmp back_jump

back_jump:
    rts

gen_colram_copy:
    lda #<colram_copy
    sta zp_temp
    lda #>colram_copy
    sta zp_temp+1
--
    ldy #5
-
    lda .vorlage,y
    sta (zp_temp),y
    dey
    bpl -

    inc .vorlage+1
    inc .vorlage+4
    lda .vorlage+1
    bne +
    inc .vorlage+2
    inc .vorlage+5
+
    clc
    lda zp_temp
    adc #6
    sta zp_temp

    lda zp_temp+1
    adc #0
    sta zp_temp+1

    lda .vorlage+4
    cmp #$e8
    bne --
    lda .vorlage+5
    cmp #$db
    bne --

    ldy #0
    lda #$60
    sta (zp_temp),y

    clc
    lda zp_temp
    adc #1
    sta zp_temp

    lda zp_temp+1
    adc #0
    sta zp_temp+1
    rts

.vorlage
    lda colram_buffer
    sta $d800

colram_copy:
