!cpu 6510
; ______________________________________________________________________
enable                = $20
disable               = $2c

irq0_line             = $11

sprmydxstart          = $03
sprmydystart          = $b9
sprmydcolor           = $01
sprmulti0             = $00
sprmulti1             = $07

; ____________ ZEROPAGE ADRESSES

dd00_value            = $05
irq_ready             = dd00_value+1
fadein_andval         = irq_ready+1
d015_value            = fadein_andval+1

; ____________ RAM ADRESSES

bank0                 = $8000
vidmem0               = bank0+$0000
spritedata0           = bank0+$0400
bitmap0               = bank0+$2000
spritept0             = vidmem0+$3f8

bank1                 = $c000
vidmem1               = bank1+$0000
spritedata1           = bank1+$0400
bitmap1               = bank1+$2000
spritept1             = vidmem1+$3f8

; ______________________________________________ INCLUDES
!source "../../bin/framework/frameworksymbols.inc"

!if RELEASE = 0 {
  *= $0400
    !bin "../../bin/framework/framework-noloader.prg",,2

  *= fw_music_init
    !bin "../framework/music/dt2remix-part-06.sid",,$7e
}
; ______________________________________________________________________

*= $1ffe
    !byte <code_start
    !byte >code_start
; ______________________________________________________________________
; ______________________________________________ IRQs
!zone IRQs
*= $2000
irq0:
    sta fw_savea
    stx fw_savex
    sty fw_savey

    lda dd00_value
    sta $dd00

    lda d015_value
    sta $d015

    lda $d012
    cmp $d012
    beq *-3

enable_sprites:
    bit sprites_setup

    lda #$00
    sta $d012

    lda #<fw_irq
    sta $fffe
    lda #>fw_irq
    sta $ffff

    lda #$01
    sta irq_ready

    asl $d019

    lda fw_savea
    ldx fw_savex
    ldy fw_savey
    rti
; ______________________________________________________________________
; ______________________________________________ CODE START
!zone MAIN
code_start:
    lda #$00
    sta dd00_value
    sta d015_value

    lda #$00
    ldx #>bitmap0
    jsr helper_bitmapfill       ; clear bitmap0

;   lda #$00
    ldx #>bitmap1
    jsr helper_bitmapfill       ; clear bitmap1

    jsr fadein_colors           ; fade in colors

    ldx #0
-
    lda source_sprites+$000,x
    sta spritedata0+$000,x      ; copy spritedata
    sta spritedata1+$000,x
    lda source_sprites+$0c0,x
    sta spritedata0+$0c0,x
    sta spritedata1+$0c0,x
    inx
    bne -

    jsr fw_wait_irq             ; set up irq0 and some VIC registers

    lda #<irq0
    sta fw_irq_lo
    lda #>irq0
    sta fw_irq_hi

    lda #$3b                    ; multicolor mode
    sta fw_d011_value

    lda #irq0_line
    sta fw_d012_value

    lda #$18                    ; multicolor mode
    sta $d016

    lda #$08
    sta $d018

fade_in:
    lda #%11000000              ; fade in bitmaps:
    sta fadein_andval           ; set value for AND
    jsr fadein_bitmap0          ; copy contents to bitmap while displaying the
    jsr wait_irq                ; other one
    lda #<!(bank0/$4000) & 3
    sta dd00_value              ; change bank switch ($dd00 set in irq0)
    ldx #$04
    jsr wait_x                  ; wait X frames

    lda #%11001100              ; .
    sta fadein_andval           ; .
    jsr fadein_bitmap1          ; .
    jsr wait_irq
    lda #<!(bank1/$4000) & 3
    sta dd00_value
    ldx #$04
    jsr wait_x

    lda #%11001111
    sta fadein_andval
    jsr fadein_bitmap0
    jsr wait_irq
    lda #<!(bank0/$4000) & 3
    sta dd00_value
    ldx #$04
    jsr wait_x

    lda #%11111111
    sta fadein_andval
    jsr fadein_bitmap1
    jsr wait_irq
    lda #<!(bank1/$4000) & 3    ; final switch
    sta dd00_value              ; current bank is now bank1

;   lda #%11111111
;   sta fadein_andval
    jsr fadein_bitmap0          ; make sure that both bitmaps are identical

    ldx #20
    jsr wait_x

    lda #enable
    sta enable_sprites          ; enable sprites

    jsr wait_irq

    jsr sprites_mydblink

    jsr wait_irq

    jsr speedcode_diff_bmp      ; prepare bitmap and screen in bank0
    jsr speedcode_diff_scr      ; for the HAPPY star

    lda #$ff
    cmp $d012
    bne *-3

    jsr speedcode_diff_col

    lda #<!(bank0/$4000) & 3
    sta dd00_value

    ldx #50
    jsr wait_x

    ldx #$00
--
    lda random_tab,x
    tay
    lda #$00
    sta vidmem0+$000,y
    sta vidmem0+$100,y
    sta vidmem0+$200,y
    sta vidmem0+$300,y
    lda #$00
    sta $d800+$000,y
    sta $d800+$100,y
    sta $d800+$200,y
    sta $d800+$300,y
    jsr wait_bottom
    inx
    bne --

    ldx #20
    jsr wait_x

exit_part:
    jsr wait_irq

    lda #<fw_irq
    sta $fffe
    sta fw_irq_lo
    lda #>fw_irq
    sta $ffff
    sta fw_irq_hi

    lda #$00
    sta $d012
    sta fw_d012_value

    lda #$0b
    sta $d011
    sta fw_d011_value
!if RELEASE = 0 {
    jmp *
} else {
    jmp fw_back2framework
}

; ==============================================================================
; memfill
; ------------+-----------------------------------------------------------------
; depends on: | -
; ------------+-----------------------------------------------------------------
; uses:       | A, X, Y
; ------------+-----------------------------------------------------------------
; preserves:  | A
; ------------+---+-------------------------------------------------------------
; input:      | A | fill value
;             | X | highbyte fill startaddress
;             | Y | highbyte fill endaddress
; ------------+---+-------------------------------------------------------------
; output:     | - |
; ------------+---+-------------------------------------------------------------
!zone MEMFILL {
helper_memfill:
    stx .from_hi
    sty .to_hi
.loop:
    ldx #0
-
.from_hi =*+2
    sta $0000,x
    inx
    bne -

    inc .from_hi
    ldy .from_hi
.to_hi =*+1
    cpy #0
    bne .loop
    rts
}
; ==============================================================================

; ==============================================================================
; bitmapfill
; ------------+-----------------------------------------------------------------
; depends on: | memfill
; ------------+-----------------------------------------------------------------
; uses:       | A, X, Y, STACK
; ------------+-----------------------------------------------------------------
; preserves:  | A
; ------------+---+-------------------------------------------------------------
; input:      | A | fill value
;             | X | highbyte bitmap startaddress
; ------------+---+-------------------------------------------------------------
; output:     | - |
; ------------+---+-------------------------------------------------------------
!zone BITMAPFILL {
helper_bitmapfill:
    pha
    txa
    clc
    adc #$1f
    sta .last_high
    tay
    pla
    jsr helper_memfill
    ldx #$3f
-
.last_high =*+2
    sta $0000,x
    dex
    bpl -
    rts
}
; ==============================================================================
; ______________________________________________________________________
!zone WAITIRQ
wait_irq:
    lda #0
    sta irq_ready
-
    lda irq_ready
    beq -
    rts

wait_x:
    jsr wait_irq
    dex
    cpx #$ff
    bne wait_x
    rts

wait_bottom:
    lda #$e0
    cmp $d012
    bne *-3
    rts
; ______________________________________________________________________
!zone BITMAP_FADES
fadein_bitmap0:
    ldx #0
-
!for i, $00, $1e {
    lda source_bitmap+(i*$100),x
    and fadein_andval
    sta bitmap0+(i*$100),x
}
    lda source_bitmap+$1e3f,x
    sta bitmap0+$1e3f,x
    inx
    beq +
    jmp -
+
    rts
fadein_bitmap1:
    ldx #0
-
!for i, $00, $1e {
    lda source_bitmap+(i*$100),x
    and fadein_andval
    sta bitmap1+(i*$100),x
}
    lda source_bitmap+$1e3f,x
    sta bitmap1+$1e3f,x
    inx
    beq +
    jmp -
+
    rts
; ______________________________________________________________________
!zone COLOR_FADE
fadein_colors:
.loop:
    ldx #1
;   jsr wait_x
.xsav =*+1
    ldx #$27
!for i, 0, 24 {
    lda source_screen+(i*40),x
    sta vidmem0+(i*40),x
    sta vidmem1+(i*40),x
    lda source_colram+(i*40),x
    sta $d800+(i*40),x
}
    dex
    bpl +
    rts
+
    stx .xsav
    jmp .loop
; ______________________________________________________________________
!zone SPRITES
sprites_setup:
!for i, 0, 7 {
    lda #<(((spritedata0-bank0)/$40)+i)
    sta spritept0+i             ; set sprite pointer
    sta spritept1+i             ; in vidmem0 / vidmem1
}

    ldx #7
    ldy #$0e
-                               ; setup
    lda .xpos_tab,x             ; x-positions
    sta $d000,y
    lda .ypos_tab,x             ; y-positions
    sta $d001,y                 ; and
    lda .color_tab,x            ; colors
    sta $d027,x
    dey
    dey
    dex
    bpl -

    lda #sprmulti0              ; setup multicolors
    sta $d025
    lda #sprmulti1
    sta $d026

    lda #0                      ; setup expand registers
    sta $d01d
    sta $d017

    lda #$ff                    ; set multicolor mode
    sta $d01c

    ldx #7
    lda #$00
-   ora .d010_tab,x             ; set $d010
    dex
    bpl -
    sta $d010

    rts

.xpos_tab:
    !byte sprmydxstart+(0*24)
    !byte sprmydxstart+(1*24)
    !byte sprmydxstart+(2*24)
    !byte $00
    !byte $00
    !byte $00
    !byte $00
    !byte $00

.ypos_tab:
    !byte sprmydystart
    !byte sprmydystart
    !byte sprmydystart
    !byte $00
    !byte $00
    !byte $00
    !byte $00
    !byte $00

.color_tab:
    !byte sprmydcolor
    !byte sprmydcolor
    !byte sprmydcolor
    !byte $00
    !byte $00
    !byte $00
    !byte $00
    !byte $00

.d010_tab:
    !byte %00000001
    !byte %00000010
    !byte %00000100
    !byte %00000000
    !byte %00000000
    !byte %00000000
    !byte %00000000
    !byte %00000000
; ______________________________________________________________________
sprites_mydblink:
    ldy #0
    lda .delay_tab,y
-
    tax
    jsr wait_x
    lda d015_value
    eor #($00 XOR $07)
    sta d015_value
    iny
    lda .delay_tab,y
    bne -
    rts

.delay_tab:
!for i, 0, $0f {
    !byte $10-i, i+1
}
    !byte $01
    !byte $00
; ______________________________________________________________________
!zone SPEEDCODES
speedcode_diff_bmp:
    lda #$00
    sta bitmap0+$01d7
    sta bitmap0+$0309
    sta bitmap0+$030a
    sta bitmap0+$030b
    sta bitmap0+$030c
    sta bitmap0+$030d
    sta bitmap0+$030e
    sta bitmap0+$030f
    sta bitmap0+$0310
    sta bitmap0+$0311
    sta bitmap0+$0312
    sta bitmap0+$0313
    sta bitmap0+$0314
    sta bitmap0+$0315
    sta bitmap0+$0316
    sta bitmap0+$0317
    sta bitmap0+$0448
    sta bitmap0+$0449
    sta bitmap0+$044a
    sta bitmap0+$044b
    sta bitmap0+$044c
    sta bitmap0+$044d
    sta bitmap0+$0450
    sta bitmap0+$0451
    sta bitmap0+$0452
    sta bitmap0+$0453
    sta bitmap0+$0d18
    lda #$03
    sta bitmap0+$0e3b
    lda #$0d
    sta bitmap0+$009c
    sta bitmap0+$009d
    sta bitmap0+$0964
    lda #$0f
    sta bitmap0+$009e
    sta bitmap0+$0446
    sta bitmap0+$0be7
    lda #$11
    sta bitmap0+$0457
    lda #$13
    sta bitmap0+$0e3a
    lda #$18
    sta bitmap0+$0e35
    lda #$1f
    sta bitmap0+$0aa9
    sta bitmap0+$0aaa
    lda #$23
    sta bitmap0+$081e
    sta bitmap0+$081f
    lda #$2f
    sta bitmap0+$0bb2
    lda #$30
    sta bitmap0+$0e39
    lda #$34
    sta bitmap0+$0e38
    lda #$35
    sta bitmap0+$0965
    sta bitmap0+$0966
    lda #$3c
    sta bitmap0+$0bb3
    lda #$3d
    sta bitmap0+$009f
    sta bitmap0+$0967
    lda #$3f
    sta bitmap0+$01d8
    sta bitmap0+$01d9
    sta bitmap0+$01da
    sta bitmap0+$01db
    sta bitmap0+$01dc
    sta bitmap0+$01dd
    sta bitmap0+$01de
    sta bitmap0+$01df
    sta bitmap0+$0318
    sta bitmap0+$0319
    sta bitmap0+$031a
    sta bitmap0+$031b
    sta bitmap0+$031c
    sta bitmap0+$031d
    sta bitmap0+$031e
    sta bitmap0+$031f
    sta bitmap0+$0458
    sta bitmap0+$0459
    sta bitmap0+$045a
    sta bitmap0+$045c
    sta bitmap0+$045d
    sta bitmap0+$045e
    sta bitmap0+$045f
    sta bitmap0+$0598
    sta bitmap0+$0599
    sta bitmap0+$059a
    sta bitmap0+$059c
    sta bitmap0+$082b
    sta bitmap0+$082e
    sta bitmap0+$082f
    sta bitmap0+$0968
    sta bitmap0+$0bb1
    sta bitmap0+$0bdf
    sta bitmap0+$0cf1
    sta bitmap0+$0d20
    lda #$40
    sta bitmap0+$0bb4
    lda #$41
    sta bitmap0+$0bb5
    lda #$4a
    sta bitmap0+$0aa0
    sta bitmap0+$0aa1
    lda #$4f
    sta bitmap0+$095d
    sta bitmap0+$096b
    sta bitmap0+$096c
    sta bitmap0+$096f
    lda #$53
    sta bitmap0+$0958
    sta bitmap0+$0959
    sta bitmap0+$0cec
    sta bitmap0+$0ced
    sta bitmap0+$0cee
    sta bitmap0+$0cef
    lda #$54
    sta bitmap0+$044e
    lda #$55
    sta bitmap0+$0326
    sta bitmap0+$044f
    sta bitmap0+$0bb6
    sta bitmap0+$0bb7
    lda #$56
    sta bitmap0+$0cff
    lda #$57
    sta bitmap0+$0321
    sta bitmap0+$0323
    sta bitmap0+$0325
    sta bitmap0+$0327
    lda #$5c
    sta bitmap0+$01e4
    sta bitmap0+$01e6
    lda #$5d
    sta bitmap0+$0324
    lda #$5f
    sta bitmap0+$0320
    sta bitmap0+$0322
    sta bitmap0+$0e23
    sta bitmap0+$0f61
    sta bitmap0+$0f63
    sta bitmap0+$0f64
    sta bitmap0+$0f65
    lda #$60
    sta bitmap0+$0ce8
    lda #$63
    sta bitmap0+$095a
    lda #$64
    sta bitmap0+$0cea
    lda #$70
    sta bitmap0+$00a7
    sta bitmap0+$01e1
    sta bitmap0+$0469
    lda #$75
    sta bitmap0+$0580
    lda #$77
    sta bitmap0+$01e7
    lda #$7c
    sta bitmap0+$01e3
    sta bitmap0+$01e5
    sta bitmap0+$046b
    sta bitmap0+$046d
    lda #$7e
    sta bitmap0+$0f60
    lda #$7f
    sta bitmap0+$082a
    sta bitmap0+$0aab
    sta bitmap0+$0aac
    sta bitmap0+$0cf2
    sta bitmap0+$0e21
    sta bitmap0+$0e22
    sta bitmap0+$0f62
    lda #$8a
    sta bitmap0+$0960
    lda #$8b
    sta bitmap0+$0961
    lda #$8f
    sta bitmap0+$0962
    sta bitmap0+$0963
    sta bitmap0+$096d
    sta bitmap0+$096e
    sta bitmap0+$0aa8
    lda #$94
    sta bitmap0+$0ceb
    lda #$98
    sta bitmap0+$0ce9
    lda #$9f
    sta bitmap0+$0e24
    sta bitmap0+$0e25
    sta bitmap0+$0e26
    lda #$a3
    sta bitmap0+$095b
    sta bitmap0+$095c
    lda #$b1
    sta bitmap0+$0baa
    lda #$bd
    sta bitmap0+$05af
    lda #$bf
    sta bitmap0+$0e20
    sta bitmap0+$0e27
    lda #$c0
    sta bitmap0+$00a5
    sta bitmap0+$00a6
    sta bitmap0+$032c
    sta bitmap0+$032d
    sta bitmap0+$032e
    sta bitmap0+$032f
    lda #$c2
    sta bitmap0+$0816
    sta bitmap0+$0e34
    lda #$c4
    sta bitmap0+$081b
    lda #$ca
    sta bitmap0+$0827
    lda #$cc
    sta bitmap0+$0f6c
    lda #$cf
    sta bitmap0+$0818
    sta bitmap0+$0bd5
    lda #$d0
    sta bitmap0+$0bd7
    lda #$d1
    sta bitmap0+$0591
    lda #$d3
    sta bitmap0+$0819
    sta bitmap0+$0bd6
    lda #$d4
    sta bitmap0+$0aa3
    lda #$d5
    sta bitmap0+$0593
    lda #$dc
    sta bitmap0+$046c
    lda #$e0
    sta bitmap0+$0826
    lda #$e1
    sta bitmap0+$0bad
    sta bitmap0+$0baf
    lda #$ec
    sta bitmap0+$0822
    lda #$f0
    sta bitmap0+$01e0
    sta bitmap0+$01e2
    sta bitmap0+$0468
    sta bitmap0+$046a
    sta bitmap0+$0821
    sta bitmap0+$0e2c
    lda #$f1
    sta bitmap0+$0956
    sta bitmap0+$0957
    sta bitmap0+$0ba9
    sta bitmap0+$0bab
    sta bitmap0+$0bac
    sta bitmap0+$0bae
    sta bitmap0+$0cf4
    sta bitmap0+$0cf5
    sta bitmap0+$0e28
    sta bitmap0+$0e29
    sta bitmap0+$0e2a
    lda #$f2
    sta bitmap0+$0e2b
    lda #$f3
    sta bitmap0+$0d12
    sta bitmap0+$0e2f
    sta bitmap0+$0f69
    sta bitmap0+$0f6b
    sta bitmap0+$0f6d
    lda #$f4
    sta bitmap0+$05a8
    sta bitmap0+$081a
    lda #$f5
    sta bitmap0+$0aa4
    sta bitmap0+$0d10
    lda #$f7
    sta bitmap0+$0a92
    sta bitmap0+$0cf3
    lda #$fc
    sta bitmap0+$0ba8
    sta bitmap0+$0cf6
    sta bitmap0+$0cf7
    sta bitmap0+$0e2d
    sta bitmap0+$0e2e
    sta bitmap0+$0e33
    sta bitmap0+$0f68
    sta bitmap0+$0f6a
    lda #$fd
    sta bitmap0+$05a3
    sta bitmap0+$05a5
    lda #$ff
    sta bitmap0+$0447
    sta bitmap0+$0460
    sta bitmap0+$0461
    sta bitmap0+$0462
    sta bitmap0+$0463
    sta bitmap0+$0464
    sta bitmap0+$0465
    sta bitmap0+$05a1
    sta bitmap0+$05a2
    sta bitmap0+$05a4
    sta bitmap0+$05a7
    sta bitmap0+$06e0
    sta bitmap0+$0823
    sta bitmap0+$0824
    sta bitmap0+$0aa5
    sta bitmap0+$0aa6
    sta bitmap0+$0aa7
    sta bitmap0+$0aad
    sta bitmap0+$0aae
    sta bitmap0+$0aaf
    sta bitmap0+$0bde
    sta bitmap0+$0be0
    sta bitmap0+$0cf0
    sta bitmap0+$0d19
    sta bitmap0+$0d1a
    sta bitmap0+$0e30
    sta bitmap0+$0e31
    sta bitmap0+$0e32
    rts
speedcode_diff_scr:
    lda #$16
    sta vidmem0+$012b
    lda #$67
    sta vidmem0+$012c
    lda #$77
    sta vidmem0+$003b
    rts
speedcode_diff_col:
    lda #$f7
    sta $d800+$003b
    rts
; ______________________________________________________________________
!zone DATA
source_bitmap:
    !bin "sadstar.bmp"
source_screen:
    !bin "sadstar.scr"
source_colram:
    !bin "sadstar.col"
source_sprites:
    !bin "mydblase.bin"

random_tab:
  !byte 208, 029, 189, 225, 145, 054, 196, 045
  !byte 091, 136, 126, 165, 236, 103, 128, 106
  !byte 019, 217, 026, 174, 032, 021, 245, 043
  !byte 147, 059, 134, 255, 206, 101, 068, 241
  !byte 016, 015, 035, 067, 239, 024, 227, 011
  !byte 040, 042, 215, 117, 139, 052, 100, 137
  !byte 243, 231, 049, 232, 234, 164, 074, 214
  !byte 194, 072, 150, 030, 143, 069, 046, 051
  !byte 033, 114, 082, 080, 075, 183, 003, 056
  !byte 159, 066, 177, 124, 209, 009, 249, 154
  !byte 031, 093, 008, 077, 251, 004, 220, 195
  !byte 211, 250, 235, 017, 071, 120, 237, 138
  !byte 254, 070, 109, 104, 187, 125, 155, 110
  !byte 179, 127, 204, 170, 178, 006, 153, 201
  !byte 229, 212, 146, 240, 084, 047, 028, 228
  !byte 115, 092, 007, 148, 210, 224, 197, 144
  !byte 246, 090, 089, 062, 001, 161, 121, 184
  !byte 157, 172, 087, 050, 002, 005, 058, 198
  !byte 036, 149, 108, 168, 248, 132, 173, 041
  !byte 095, 202, 158, 186, 166, 130, 116, 238
  !byte 203, 111, 076, 044, 020, 140, 012, 023
  !byte 105, 160, 099, 112, 247, 022, 083, 034
  !byte 135, 025, 244, 096, 185, 060, 053, 102
  !byte 061, 167, 233, 057, 131, 119, 190, 085
  !byte 223, 252, 182, 129, 107, 226, 013, 253
  !byte 133, 010, 200, 169, 063, 181, 175, 221
  !byte 207, 171, 037, 141, 205, 098, 018, 055
  !byte 113, 027, 222, 219, 065, 152, 156, 039
  !byte 123, 142, 038, 081, 199, 162, 216, 078
  !byte 192, 118, 191, 014, 094, 230, 218, 073
  !byte 048, 122, 088, 242, 188, 176, 086, 163
  !byte 000, 180, 213, 079, 151, 064, 097, 193
